import logging
import os
import sys
from typing import Optional

import rosys
from nicegui import ui

from ..event import Event
from .communication import Communication, SerialCommunication


class RobotBrain:
    '''This module manages the communication with a [Zauberzeug Robot Brain](https://zauberzeug.com/robot-brain.html).

    It expects a communication object, which is used for the actual read and write operations.
    Besides providing some basic methods like configuring or restarting the microcontroller, it augments and verifies checksums for each message.
    '''

    def __init__(self, communication: Communication, lizard_startup: str = 'lizard.txt') -> None:
        self.LINE_RECEIVED = Event()
        '''a line has been received from the microcontroller (argument: line as string)'''

        self.communication = communication
        self.lizard_startup = lizard_startup
        self.waiting_list: dict[str, Optional[str]] = {}

        self.clock_offset: Optional[float] = None
        self.hardware_time: Optional[float] = None
        self.flash_params: list[str] = []
        self.log = logging.getLogger('rosys.robot_rain')

    async def configure(self) -> None:
        self.log.info('Configuring...')
        await self.send(f'!-')
        with open(self.lizard_startup) as f:
            for line in f.read().splitlines():
                await self.send(f'!+{line}')
        await self.send(f'!.')
        await self.restart()

    async def restart(self) -> None:
        await self.send(f'core.restart()')

    async def read_lines(self) -> list[tuple[float, str]]:
        lines: list[tuple[float, str]] = []
        millis = None
        while True:
            unchecked = await self.communication.read()
            line = check(unchecked)
            if not line:
                break
            words = line.split()
            if not words:
                continue
            first = words.pop(0)
            if first in self.waiting_list:
                self.waiting_list[first] = line
            if first == 'core':
                millis = float(words.pop(0))
                if self.clock_offset is None:
                    continue
                self.hardware_time = millis / 1000 + self.clock_offset
            self.LINE_RECEIVED.emit(line)
            lines.append((self.hardware_time, line))
        if millis is not None:
            self.clock_offset = rosys.time() - millis / 1000
        return lines

    async def send(self, msg: str) -> None:
        await self.communication.send(augment(msg))

    async def send_and_await(self, msg: str, ack: str, *, timeout: float = float('inf')) -> Optional[str]:
        self.waiting_list[ack] = None
        await self.send(msg)
        t0 = rosys.time()
        while self.waiting_list.get(ack) is None and rosys.time() < t0 + timeout:
            await rosys.sleep(0.1)
        return self.waiting_list.pop(ack) if ack in self.waiting_list else None

    async def flash(self) -> None:
        with ui.dialog() as dialog, ui.card():
            status = ui.markdown('.... flashing ....')
        dialog.open()
        # NOTE this is a workaround for a bug in the current version of NiceGUI, see https://github.com/zauberzeug/nicegui/issues/131
        rosys.task_logger.create_task(dialog.page.update())
        assert isinstance(self.communication, SerialCommunication)
        self.communication.disconnect()
        output = await rosys.run.sh(['./flash.py'] + self.flash_params, timeout=None, working_dir=os.path.expanduser('~/.lizard'))
        status.set_content(f'```\n{output}\n```')
        self.communication.connect()

    def developer_ui(self) -> None:
        ui.label('Lizard')
        ui.button('Configure', on_click=self.configure).props('outline')
        ui.button('Restart', on_click=self.restart).props('outline')
        ui.button('Flash', on_click=self.flash).props('outline')
        ui.button('Enable', on_click=self.enable_esp).props('outline')

    def enable_esp(self) -> None:
        try:
            sys.path.insert(1, os.path.expanduser('~/.lizard'))
            from esp import Esp
            esp = Esp()
            with esp.pin_config():
                esp.activate()
        except:
            self.log.exception('Could not enable ESP')

    def __del__(self) -> None:
        self.communication.disconnect()

    def __repr__(self) -> str:
        return f'<RobotBrain {self.communication}>'


def augment(line: str) -> str:
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    return f'{line}@{checksum:02x}'


def check(line: Optional[str]) -> str:
    if line is None:
        return ""
    if line[-3:-2] == '@':
        check = int(line[-2:], 16)
        line = line[:-3]
        checksum = 0
        for c in line:
            checksum ^= ord(c)
        if checksum != check:
            return None
    return line



from itertools import groupby

import humanize
from matplotlib.colors import hsv_to_rgb, to_hex
from nicegui import ui

from .kpi_buckets import Month, TimeBucket, Week
from .kpi_logger import KpiLogger, str_to_date


class KpiPage:

    def __init__(self,
                 kpi_logger: KpiLogger,
                 *,
                 timespans: dict[int, str],
                 positives: dict[str, str],
                 negatives: dict[str, str],
                 title: str = '',
                 positive_title: str = '',
                 negative_title: str = '') -> None:
        self.kpi_logger = kpi_logger
        self.positives = positives
        self.negatives = negatives

        humanize.activate('de')

        @ui.page('/kpis')
        def page():
            with ui.row().style('margin:1em'):
                ui.markdown(f'### {title}').style('margin:1.5em;margin-top:-1.2em;')
                toggle = ui.toggle(timespans, value=list(timespans.keys())[0], on_change=lambda e: self.show(e.value))
            with ui.row():
                self.positive_chart = ui.chart({
                    'title': {'text': positive_title},
                    'chart': {'type': 'column'},
                    'yAxis': {'title': False},
                    'credits': False,
                    'exporting': False,
                })
                self.negative_chart = ui.chart({
                    'title': {'text': negative_title},
                    'chart': {'type': 'column'},
                    'plotOptions': {'series': {'stacking': 'normal'}},
                    'yAxis': {'title': False},
                    'colors': [to_hex(hsv_to_rgb((0, 1, i / 100.0))) for i in reversed(range(20, 120, 20))],
                    'credits': False,
                    'exporting': False,
                })
            ui.timer(5, lambda: self.show(toggle.value))

    def show(self, num_days: int) -> None:
        if num_days <= 7:
            time_buckets = self.kpi_logger.days[-num_days:]
        elif num_days % 7 == 0:
            weeks = [list(v) for _, v in groupby(self.kpi_logger.days,
                                                 key=lambda d: str_to_date(d.date).strftime('%y-%W'))][-num_days//7:]
            time_buckets = [Week.from_buckets(days_in_week) for days_in_week in weeks]
        elif num_days % 30 == 0 and num_days <= 90:
            months = [list(v) for _, v in groupby(self.kpi_logger.days,
                                                  key=lambda d: str_to_date(d.date).strftime('%y-%m'))][-num_days//30:]
            time_buckets = [Month.from_buckets(days_in_month) for days_in_month in months]
        else:
            raise ValueError(f'Unsupported number of days: {num_days}')

        keys = set(key for day in time_buckets for key in day.incidents if key in self.positives)
        data = {self.positives[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
        self.positive_chart.options.xAxis.categories = [self._label(b) for b in time_buckets]
        self.positive_chart.options.series = [{'name': k, 'data': v} for k, v in sorted(data.items())]
        self.positive_chart.update()

        keys = set(key for day in time_buckets for key in day.incidents if key in self.negatives)
        data = {self.negatives[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
        self.negative_chart.options.xAxis.categories = [self._label(b) for b in time_buckets]
        self.negative_chart.options.series = [{'name': k, 'data': v} for k, v in sorted(data.items())]
        self.negative_chart.update()

    @staticmethod
    def _label(time_bucket: TimeBucket) -> str:
        if isinstance(time_bucket, Month):
            return str_to_date(time_bucket.date).strftime('%b')
        elif isinstance(time_bucket, Week):
            return f'#{str_to_date(time_bucket.date).isocalendar()[1]}'
        else:
            return humanize.naturaldate(str_to_date(time_bucket.date))



import abc
from typing import Optional

from aenum import Enum, auto

from ..event import Event
from .detections import Detections
from .image import Image
from .uploads import Uploads


class Autoupload(Enum, init='value __doc__'):
    '''Configures the auto-submitting of images to the Learning Loop'''

    def _generate_next_value_(name: str, start, count, last_values) -> str:
        '''uses enum name as value when calling auto()'''
        return name.lower()

    FILTERED = auto(), 'only submit images with novel detections and in an uncertainty range (this is the default)'
    DISABLED = auto(), 'no auto-submitting'
    ALL = auto(), 'submit all images which are run through the detector'


class Detector(abc.ABC):
    '''A detector allows detecting objects in images.

    It also holds an upload queue for sending images with uncertain results to an active learning infrastructure like the [Zauberzeug Learning Loop](https://zauberzeug.com/learning-loop.html).
    '''

    def __init__(self) -> None:
        self.NEW_DETECTIONS = Event()
        '''detection on an image is completed (argument: image)'''

    @abc.abstractmethod
    async def detect(self, image: Image, autoupload: Autoupload = Autoupload.FILTERED) -> Optional[Detections]:
        return

    @property
    def uploads(self) -> Uploads:
        return



from functools import wraps
from typing import Optional

from nicegui import ui


class Track:

    def __init__(self) -> None:
        self.stack: list[str] = []
        self._ui: Optional[ui.label] = None

    def __call__(self, f):
        @wraps(f)
        async def wrap(*args, **kw):
            try:
                self.stack.append(f.__name__)
                return await f(*args, **kw)
            finally:
                if self.stack and self.stack[-1] == f.__name__:
                    self.stack.pop()
        return wrap

    def ui(self) -> ui.label:
        def update() -> None:
            self._ui.text = ' → '.join(self.stack)
        self._ui = ui.label()
        ui.timer(0.5, update)
        return self._ui

    def reset(self) -> None:
        self.stack.clear()
        if self._ui:
            self._ui.text = ''


track = Track()



from typing import Optional

import rosys

from ..geometry import Pose, PoseStep, Velocity
from .wheels import Wheels


class WheelsSimulation(Wheels):
    '''This module simulates two wheels.

    Drive and stop commands impact internal velocities (linear and angular).
    A simulated pose is regularly updated with these velocities, while the velocities are emitted as an event.
    '''

    def __init__(self) -> None:
        super().__init__()

        self.pose: Pose = Pose()
        self.linear_velocity: float = 0
        self.angular_velocity: float = 0

        rosys.on_repeat(self.step, 0.01)
        self._last_step: Optional[float] = None

    async def drive(self, linear: float, angular: float) -> None:
        self.linear_velocity = linear
        self.angular_velocity = angular

    async def stop(self) -> None:
        self.linear_velocity = 0
        self.angular_velocity = 0

    async def step(self) -> None:
        if self._last_step is not None:
            dt = rosys.time() - self._last_step
            self.pose += PoseStep(linear=dt*self.linear_velocity, angular=dt*self.angular_velocity, time=rosys.time())
        self._last_step = rosys.time()
        velocity = Velocity(linear=self.linear_velocity, angular=self.angular_velocity, time=rosys.time())
        self.VELOCITY_MEASURED.emit([velocity])



import logging
from typing import TYPE_CHECKING, Optional

import cv2
import numpy as np
from nicegui import ui
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route

from .. import run
from .camera import Camera
from .image import Image

if TYPE_CHECKING:
    from .camera_provider import CameraProvider

log = logging.getLogger('rosys.image_route')


def create_image_route(camera_provider: 'CameraProvider') -> None:
    async def get_image(request: Request, **_) -> Response:
        return await _get_image(camera_provider.cameras, request)
    ui.add_route(Route('/' + camera_provider.base_path + '/{camera_id}/{timestamp}', get_image))
    ui.add_route(Route('/' + camera_provider.base_path + '/placeholder', _get_placeholder))


def _get_placeholder(_) -> Response:
    return Response(content=Image.create_placeholder('no image').data, media_type='image/jpeg')


async def _get_image(cameras: dict[str, Camera], request: Request) -> None:
    try:
        camera = cameras.get(request.path_params['camera_id'])
        if not camera:
            return Response(content='Camera not found', status_code=404)

        shrink_factor = int(request.query_params.get('shrink', 1))
        jpeg = await _try_get_jpeg(camera, request.path_params['timestamp'], shrink_factor)
        if not jpeg:
            return Response(content='Image not found', status_code=404)

        return Response(content=jpeg, headers={'cache-control': 'max-age=7776000'}, media_type='image/jpeg')
    except:
        log.exception('could not get image')
        raise


async def _try_get_jpeg(camera: Camera, timestamp: str, shrink_factor: int) -> Optional[bytes]:
    for image in reversed(camera.images):
        if str(image.time) == timestamp and image.data is not None:
            jpeg = image.data
            if shrink_factor != 1:
                array = np.frombuffer(image.data, dtype=np.uint8)
                if array is None:
                    return
                jpeg = await run.cpu_bound(_shrink, shrink_factor, array)
            return jpeg


def _shrink(factor: int, array: np.ndarray) -> bytes:
    decoded = cv2.imdecode(array, cv2.IMREAD_COLOR)
    if decoded is None:
        return
    img = decoded[::factor, ::factor]
    return cv2.imencode('.jpg', img, [int(cv2.IMWRITE_JPEG_QUALITY), 60])[1].tobytes()



from .communication import Communication, SerialCommunication, WebCommunication
from .robot_brain import RobotBrain
from .wheels import Wheels
from .wheels_hardware import WheelsHardware
from .wheels_simulation import WheelsSimulation



import asyncio
import logging
from typing import Callable, Coroutine, Optional

import rosys
from rosys.analysis import track

from ..driving import Drivable, Steerer
from ..event import Event
from ..task_logger import create_task
from .automation import Automation


class Automator:
    '''An automator allows running automations, i.e. coroutines that can be paused and resumed.

    See [Click-and-drive](https://rosys.io/examples/click-and-drive/) for a simple example of an automation.

    _wheels_: Optional wheels (or any stoppable hardware representation) will be stopped when an automation pauses or stops.

    _steerer_: If provided, manually steering the robot will pause a currently running automation.

    _default_automation_: If provided, it allows the automator to start a new automation without passing an automation 
    (e.g. via an "Play"-button like offered by the [automation controls](https://rosys.io/reference/rosys/automation/#rosys.automation.automation_controls)). 
    The passed function should return a new coroutine on every call.
    '''

    def __init__(self,
                 wheels: Optional[Drivable],
                 steerer: Optional[Steerer], *,
                 default_automation: Optional[Callable] = None) -> None:
        self.AUTOMATION_STARTED = Event()
        '''an automation has been started'''

        self.AUTOMATION_PAUSED = Event()
        '''an automation has been paused (string argument: description of the cause)'''

        self.AUTOMATION_RESUMED = Event()
        '''an automation has been resumed'''

        self.AUTOMATION_STOPPED = Event()
        '''an automation has been stopped (string argument: description of the cause)'''

        self.AUTOMATION_FAILED = Event()
        '''an automation has failed to complete (string argument: description of the cause)'''

        self.AUTOMATION_COMPLETED = Event()
        '''an automation has been completed'''

        self.log = logging.getLogger('rosys.automator')

        self.default_automation = default_automation

        self.enabled: bool = True
        self.automation: Optional[Automation] = None

        if steerer:
            steerer.STEERING_STARTED.register(lambda: self.pause(because='steering started'))

        if wheels:
            self.AUTOMATION_PAUSED.register(lambda _: create_task(wheels.stop()))
            self.AUTOMATION_STOPPED.register(lambda _: create_task(wheels.stop()))

        rosys.on_shutdown(lambda: self.stop(because='automator is shutting down'))

    @property
    def is_stopped(self) -> bool:
        return self.automation is None or self.automation.is_stopped

    @property
    def is_running(self) -> bool:
        return self.automation is not None and self.automation.is_running

    @property
    def is_paused(self) -> bool:
        return self.automation is not None and self.automation.is_paused

    def start(self, coro: Optional[Coroutine] = None) -> None:
        '''Starts a new automation.

        You can pass any coroutine.
        The automator will make sure it can be paused, resumed and stopped.'''
        if coro is None:
            self.start(self.default_automation())
            return
        if not self.enabled:
            coro.close()
            return
        self.stop(because='new automation starts')
        self.automation = Automation(coro, self._handle_exception, on_complete=self._on_complete)
        create_task(asyncio.wait([self.automation]), name='automation')
        self.AUTOMATION_STARTED.emit()
        rosys.notify('automation started')

    def pause(self, because: str) -> None:
        '''Pauses the current automation.

        You need to provide a cause as a meaningful string which will be used as notification.
        '''

        if self.is_running:
            self.automation.pause()
            self.AUTOMATION_PAUSED.emit(because)
            rosys.notify(f'automation paused because {because}')

    def resume(self) -> None:
        '''Resumes the current automation.'''

        if not self.enabled:
            return
        if self.is_paused:
            self.automation.resume()
            self.AUTOMATION_RESUMED.emit()
            rosys.notify('automation resumed')

    def stop(self, because: str) -> None:
        '''Stops the current automation.

        You need to provide a cause as a meaningful string which will be used as notification
        '''

        if not self.is_stopped:
            self.automation.stop()
            self.AUTOMATION_STOPPED.emit(because)
            track.reset()
            rosys.notify(f'automation stopped because {because}')

    def enable(self) -> None:
        '''Enables the automator.

        It is enabled by default. It can be disabled by calling `disable()`.'''

        self.enabled = True

    def disable(self, because: str) -> None:
        '''Disables the automator.

        No automations can be started while the automator is disabled.
        If an automation is running or paused it will be stopped.
        You need to provide a cause of the disabling as a meaningful string. 
        This will be used as notification message.
        '''

        self.stop(because)
        self.enabled = False

    def _handle_exception(self, e: Exception) -> None:
        self.stop(because='an exception occurred in an automation')
        self.AUTOMATION_FAILED.emit(str(e))
        rosys.notify('automation failed')
        if rosys.is_test:
            self.log.exception('automation failed', e)

    def _on_complete(self) -> None:
        self.AUTOMATION_COMPLETED.emit()
        rosys.notify('automation completed')



from nicegui import ui
from rosys import config

from .automator import Automator


class AutomationControls:
    '''This UI element contains start/stop/pause/resume buttons for controlling a given automator.'''

    def __init__(self, automator: Automator) -> None:
        play_button = ui.button(on_click=lambda: automator.start()) \
            .props('icon=play_arrow unelevated').tooltip('start automation')
        pause_button = ui.button(on_click=lambda: automator.pause(because='pause button was pressed')) \
            .props('icon=pause outline').tooltip('stop automation')
        resume_button = ui.button(on_click=lambda: automator.resume()) \
            .props('icon=play_arrow outline').tooltip('resume automation')
        stop_button = ui.button(on_click=lambda: automator.stop(because='stop button was pressed')) \
            .props('icon=stop outline').tooltip('stop automation')

        def refresh() -> None:
            play_button.visible = automator.is_stopped
            pause_button.visible = automator.is_running
            resume_button.visible = automator.is_paused
            self._disable(play_button, automator.default_automation is None or not automator.enabled)
            self._disable(stop_button, automator.is_stopped)

        ui.timer(config.ui_update_interval, refresh)

    def _disable(self, button: ui.button, should_disable: bool) -> None:
        if getattr(button.view, 'disable', False) != should_disable:
            button.view.disable = should_disable
            button.update()



import logging
from datetime import datetime, timedelta
from typing import Any, Callable, Optional

import rosys
import suntime
from nicegui import ui
from rosys import persistence

from .automator import Automator

DAYS = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']


class Schedule:

    def __init__(self,
                 automator: Automator, *,
                 on_activate: Optional[Callable] = None,
                 on_deactivate: Optional[Callable] = None,
                 location: Optional[tuple[float, float]] = None,
                 locations: Optional[dict[tuple[float, float], str]] = None,
                 sunrise_offset: float = 0.0,
                 sunset_offset: float = 0.0,
                 is_enabled: bool = False,
                 ) -> None:
        '''Schedules automations according to a time plan.

        param on_activate: automation to execute when entering an active period
        param on_deactivate: automation to execute when entering an inactive period
        param location: optional geographic coordinates to activate the robot at daylight only
        param locations: optional dictionary with geographic coordinates to choose from
        param sunrise_offset: optional offset for activation after sunrise (minutes, default: 0.0)
        param sunset_offset: optional offset for deactivation after sunset (minutes, default: 0.0)
        param is_enabled: whether the schedule is enabled (default: False)
        '''
        self.log = logging.getLogger('rosys.schedule')

        self.automator = automator
        self.on_activate = on_activate
        self.on_deactivate = on_deactivate
        self.location = location
        self.locations = locations
        self.sunrise_offset = sunrise_offset
        self.sunset_offset = sunset_offset
        self.is_enabled = is_enabled

        self.half_hours = [True] * 2 * 24 * 7

        self._is_active = False
        self._sun_start_hour = 0.0
        self._sun_stop_hour = 24.0
        self.buttons: list[tuple(ui.button, ui.button, ui.button)] = []
        rosys.on_repeat(self.step, 1)
        rosys.on_repeat(self.update_ui, 60)  # NOTE: to update the "now" indicator

        self.needs_backup: bool = False
        persistence.register(self)

    def backup(self) -> dict:
        return {
            'location': self.location,
            'sunrise_offset': self.sunrise_offset,
            'sunset_offset': self.sunset_offset,
            'is_enabled': self.is_enabled,
            'is_active': self._is_active,
            'half_hours': self.half_hours,
        }

    def restore(self, data: dict[str, Any]) -> None:
        self.location = tuple(data.get('location')) if data.get('location') else None
        self.sunrise_offset = data.get('sunrise_offset')
        self.sunset_offset = data.get('sunset_offset')
        self.is_enabled = data.get('is_enabled', False)
        self._is_active = data.get('is_active', False)
        self.half_hours[:] = data.get('half_hours', True)
        self.update_ui()

    def invalidate(self) -> None:
        self.needs_backup = True
        self.update_ui()

    def time_to_index(self, weekday: int, hour: int, minute: int) -> int:
        if not 0 <= weekday <= 6:
            raise ValueError(f'day of week must be between 0 and 6, not {weekday}')
        if not 0 <= minute <= 60:
            raise ValueError(f'minutes must be between 0 and 60, not {minute}')
        return weekday * 2 * 24 + hour * 2 + minute // 30

    def is_dark(self, hour: int, minute: int) -> bool:
        return not self._sun_start_hour < hour + minute / 60 < self._sun_stop_hour

    def is_planned(self, weekday: int, hour: int, minute: Optional[int] = None) -> bool:
        if minute is None:
            return self.half_hours[self.time_to_index(weekday, hour, 0)] \
                or self.half_hours[self.time_to_index(weekday, hour, 30)]
        else:
            return self.half_hours[self.time_to_index(weekday, hour, minute)]

    def fill(self, value: bool) -> None:
        for i in range(len(self.half_hours)):
            self.half_hours[i] = value

    def can_be_active(self) -> bool:
        now = datetime.fromtimestamp(rosys.time())
        return self.is_planned(now.weekday(), now.hour, now.minute) and not self.is_dark(now.hour, now.minute)

    def step(self) -> None:
        if not self.is_enabled:
            return
        if not self._is_active and self.can_be_active():
            self.log.info('activate')
            if self.on_activate:
                self.log.info('start automation')
                self.automator.start(self.on_activate())
            self._is_active = True
        if self._is_active and not self.can_be_active():
            self.log.info('deactivate')
            if self.on_deactivate:
                self.log.info('start automation')
                self.automator.start(self.on_deactivate())
            self._is_active = False

    def ui(self) -> ui.row:
        with ui.column() as grid:
            with ui.row().classes('fit items-center justify-between'):
                ui.switch('enabled').bind_value(self, 'is_enabled')
                if self.locations:
                    ui.select(self.locations, label='Location', on_change=lambda e: self.set_location(e.value)) \
                        .bind_value(self, 'location').style('width: 21em')
                    ui.number('Sunrise offset', format='%.0f', on_change=self.invalidate) \
                        .bind_value(self, 'sunrise_offset').props('suffix=min').style('width:100px')
                    ui.number('Sunset offset', format='%.0f', on_change=self.invalidate) \
                        .bind_value(self, 'sunset_offset').props('suffix=min').style('width:100px')

            with ui.column().style('gap: 0.3em'):
                for d in range(7):
                    with ui.row().style('gap: 0.3em'):
                        ui.label(DAYS[d]).classes('mt-2').style('width: 2em')
                        for h in range(24):
                            with ui.column().style('gap: 0.1em'):
                                hour = ui.button(f'{h:02d}', on_click=lambda _, d=d, h=h: self._toggle(d, h)) \
                                    .props('unelevated dense')
                                with ui.row().style('gap: 0.1em'):
                                    first_half = ui.button(on_click=lambda _, d=d, h=h: self._toggle(d, h, 0)) \
                                        .props('unelevated dense')
                                    second_half = ui.button(on_click=lambda _, d=d, h=h: self._toggle(d, h, 30)) \
                                        .props('unelevated dense')
                            self.buttons.append((hour, first_half, second_half))

        self.update_ui()
        return grid

    def set_location(self, location: tuple[float, float]) -> None:
        self.location = location
        self.update_ui()

    def update_sun_limits(self) -> None:
        if self.location:
            sun = suntime.Sun(lat=self.location[0], lon=self.location[1])
            start_time = sun.get_local_sunrise_time() + timedelta(minutes=self.sunrise_offset)
            stop_time = sun.get_local_sunset_time() + timedelta(minutes=self.sunset_offset)
            self._sun_start_hour = start_time.hour + start_time.minute / 60 + start_time.second / 60 / 60
            self._sun_stop_hour = stop_time.hour + stop_time.minute / 60 + stop_time.second / 60 / 60
        else:
            self._sun_start_hour = 0.0
            self._sun_stop_hour = 24.0

    def update_ui(self) -> None:
        def color(weekday: int, hour: int, minutes: list[int]) -> str:
            # https://maketintsandshades.com/#21ba45,c10015
            if all(self.is_dark(hour, minute) for minute in minutes):
                return '#d3f1da' if self.is_planned(weekday, hour, minutes[0]) else '#f3ccd0'
            dt = datetime.fromtimestamp(rosys.time())
            is_now = dt.weekday() == weekday and dt.hour == hour
            positive = '#147029' if is_now else '#21ba45' if d < 5 else '#1a9537'
            negative = '#74000d' if is_now else '#c10015' if d < 5 else '#9a0011'
            return positive if self.is_planned(weekday, hour, minutes[0]) else negative
        self.update_sun_limits()
        if not self.buttons:
            return
        for d in range(7):
            for h in range(24):
                buttons: tuple[ui.button] = self.buttons[d * 24 + h]
                styles = [
                    f'background-color: {color(d, h, [0, 30, 60])} !important',
                    f'background-color: {color(d, h, [0, 30])} !important; height: 1em; width: 0.8em',
                    f'background-color: {color(d, h, [30, 60])} !important; height: 1em; width: 0.8em',
                ]
                for i in range(3):
                    if buttons[i].view.style.lstrip(';') != styles[i]:
                        buttons[i].style(styles[i])

    def _toggle(self, weekday: int, hour: int, minute: Optional[int] = None) -> None:
        new_value = not self.is_planned(weekday, hour, minute)
        if minute is None:
            self.half_hours[self.time_to_index(weekday, hour, 0)] = new_value
            self.half_hours[self.time_to_index(weekday, hour, 30)] = new_value
        else:
            self.half_hours[self.time_to_index(weekday, hour, minute)] = new_value
        self.needs_backup = True
        self.update_ui()



import abc
import logging

import rosys

from ..event import Event


class Wheels(abc.ABC):
    '''The wheels module is a simple example for a representation of real or simulated robot hardware.

    Wheels can be moved using the `drive` methods and provide measured velocities as an event.
    '''

    def __init__(self) -> None:
        self.VELOCITY_MEASURED = Event()
        '''new velocity measurements are available for processing (argument: list of velocities)'''

        self.log = logging.getLogger(__name__)

        rosys.on_shutdown(self.stop)

    @abc.abstractmethod
    async def drive(self, linear: float, angular: float) -> None:
        pass

    @abc.abstractmethod
    async def stop(self) -> None:
        pass



import os
from glob import glob
from pathlib import Path

from nicegui import ui
from starlette import responses


class VideosPage:

    def __init__(self) -> None:
        @ui.page('/videos', 'Videos')
        def page():
            self.list = ui.column()
            ui.timer(5, self.update_list)

    def update_list(self) -> None:
        self.list.clear()
        with self.list:
            for mp4 in sorted(glob(os.path.expanduser('~/.rosys/timelapse/videos/*mp4')), reverse=True):
                name = Path(mp4).stem
                with ui.page(f'/videos/{name}', name) as page:
                    ui.html(self.create_video_tag(name))
                ui.link(name.replace('_', ' ').replace('-', ':'), page)

    @ui.get('/timelapse/{name}')
    def produce_plain_response(name: str) -> responses.FileResponse:
        return responses.FileResponse(os.path.expanduser(f'~/.rosys/timelapse/videos/{name}.mp4'))

    def create_video_tag(self, name: str) -> str:
        return f'''
<video style="
  position: absolute;
  left: 50%;
  top: 50%;
  max-height: 100%;
  max-width: 100%;
  transform: translate(-50%, -50%);
" autoplay muted loop controls>
  <source src="timelapse/{name}" type="video/mp4">
Your browser does not support the video tag.
</video>
'''



import rosys

from ..geometry import Velocity
from .robot_brain import RobotBrain
from .wheels import Wheels


class WheelsHardware(Wheels):
    '''This module implements wheels hardware.

    Drive and stop commands are forwarded to a given Robot Brain.
    Velocities are read and emitted regularly.
    '''

    def __init__(self, robot_brain: RobotBrain) -> None:
        super().__init__()

        self.robot_brain = robot_brain
        rosys.on_repeat(self.step, 0.01)
        rosys.on_shutdown(self.stop)

    async def drive(self, linear: float, angular: float) -> None:
        await self.robot_brain.send(f'wheels.speed({linear}, {angular})')

    async def stop(self) -> None:
        await self.robot_brain.send('wheels.off()')

    async def step(self) -> None:
        velocities: list[Velocity] = []
        for time, line in await self.robot_brain.read_lines():
            words = line.split()
            if words[0] == 'core':
                velocities.append(Velocity(linear=float(words[2]), angular=float(words[3]), time=time))
                await self.on_core_output(words[4:])
        self.VELOCITY_MEASURED.emit(velocities)

    async def on_core_output(self, words: list[str]) -> None:
        pass



'''source: <https://quantlane.com/blog/ensure-asyncio-task-exceptions-get-logged/>'''

import asyncio
import functools
import logging
from typing import Any, Awaitable, Optional, TypeVar

T = TypeVar('T')

logger = logging.getLogger(__name__)


def create_task(coroutine: Awaitable[T], *,
                loop: Optional[asyncio.AbstractEventLoop] = None, name: str = 'unnamed task') -> asyncio.Task[T]:
    '''
    This helper function wraps a ``loop.create_task(coroutine())`` call and ensures there is
    an exception handler added to the resulting task. If the task raises an exception it is logged
    using the provided ``logger``, with additional context provided by ``message`` and optionally
    ``message_args``.
    '''
    message = 'Task raised an exception'
    message_args = ()
    if loop is None:
        loop = asyncio.get_running_loop()
    task = loop.create_task(coroutine, name=name)
    task.add_done_callback(functools.partial(_handle_task_result, message=message, message_args=message_args))
    return task


def _handle_task_result(task: asyncio.Task, *, message: str, message_args: tuple[Any, ...] = ()) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    # Ad the pylint ignore: we want to handle all exceptions here so that the result of the task
    # is properly logged. There is no point re-raising the exception in this callback.
    except Exception:  # pylint: disable=broad-except
        logger.exception(message, *message_args)
